#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <math.h>

#include "./game_of_life_lib.h"

bool* bool_matrix_at_wrap(BoolMatrix* m, int x, int y) {
  int w = bool_matrix_width(m);
  int h = bool_matrix_height(m);
  x = x % w;
  y = y % h;
  if (x < 0) {
    x += w;
  }
  if (y < 0) {
    y += h;
  }
  return bool_matrix_at(m, x, y);
}

size_t count_alive_neighbors(BoolMatrix* m, int x, int y) {
  size_t n = 0;
  for (int x1 = ((int) x) - 1; x1 <= x + 1; x1++) {
    for (int y1 = ((int) y) - 1; y1 <= y + 1; y1++) {
      if (!(x1 == x && y1 == y)) {
        if (*bool_matrix_at_wrap(m, x1, y1)) {
          n++;
        }
      }
    }
  }
  return n;
}

void simulate(BoolMatrix* m_cur, BoolMatrix* m_next) {
  for (size_t x = 0; x < bool_matrix_width(m_cur); ++x) {
    for (size_t y = 0; y < bool_matrix_height(m_cur); ++y) {
      size_t n = count_alive_neighbors(m_cur, x, y);
      bool b_cur = *bool_matrix_at(m_cur, x, y);
      bool* b_next = bool_matrix_at(m_next, x, y);
      if (b_cur) {
        *b_next = n == 2 || n == 3;
      } else {
        *b_next = n == 3;
      }
    }
  }
}

#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))

Size2 min2(Size2 v, Size2 w) {
  return (Size2) { MIN(v.x, w.x), MIN(v.y, w.y) };
}

Size2 max2(Size2 v, Size2 w) {
  return (Size2) { MAX(v.x, w.x), MAX(v.y, w.y) };
}

void save_matrix(BoolMatrix* m, const char* file_path, Size2 begin, Size2 end) {
  Size2 min = min2(begin, end);
  Size2 max = max2(begin, end);
  FILE* f = fopen(file_path, "w");
  if (f == NULL) {
    return;
  }
  fprintf(f, "%ld,%ld\n", max.x - min.x, max.y - min.y);
  for (size_t y = min.y; y < max.y; ++y) {
    for (size_t x = min.x; x < max.x; ++x) {
      bool b = *bool_matrix_at(m, x, y);
      fprintf(f, "%c", b ? '#' : '.');
    }
    fprintf(f, "\n");
  }
  fclose(f);
}

BoolMatrix* load_matrix(const char* file_path) {
  FILE* f = fopen(file_path, "r");
  if (f == NULL) {
    return NULL;
  }

  size_t width = 0;
  size_t height = 0;
  int res = fscanf(f, "%ld,%ld\n", &width, &height);
  if (res != 2) {
    return NULL;
  }

  BoolMatrix* m = bool_matrix_new(width, height, false);
  size_t x = 0;
  size_t y = 0;
  while (true) {
    int c = fgetc(f);
    if (c == EOF) {
      break;
    }
    else if (c == '\n') {
      x = 0;
      y++;
    }
    else if (c == '#') {
      *bool_matrix_at(m, x++, y) = true;
    } else if (c == '.') {
      *bool_matrix_at(m, x++, y) = false;
    } else if (c == '.') {
      bool_matrix_free(m);
      return NULL;
    }
  }
  return m;
}

void swap(BoolMatrix** m1, BoolMatrix** m2) {
  BoolMatrix* tmp = *m1;
  *m1 = *m2;
  *m2 = tmp;
}

void save_selection(GameState* gs, const char* file_path) {
  if (gs->selection_active) {
    gs->selection_active = false;
    Size2 selection_end = (Size2) {gs->cursor_pos.x + 1, gs->cursor_pos.y + 1};
    save_matrix(gs->m_cur, file_path, gs->selection_start, selection_end);
  } else {
    save_matrix(gs->m_cur, file_path, (Size2) {0, 0}, gs->field_size);
  }
}

void load_selection(GameState* gs, const char* file_path) {
  BoolMatrix* m = load_matrix(file_path);
  if (m != NULL) {
    for (size_t x = 0; x < bool_matrix_width(m); ++x) {
      for (size_t y = 0; y < bool_matrix_height(m); ++y) {
        size_t x2 = x + gs->cursor_pos.x;
        size_t y2 = y + gs->cursor_pos.y;
        if (x2 < gs->field_size.x && y2 < gs->field_size.y) {
          *bool_matrix_at(gs->m_cur, x2, y2) = *bool_matrix_at(m, x, y);
        }
      }
    }
    bool_matrix_free(m);
  }
}

bool handle_input(GameState* gs, char c) {
  if (c == 'q') {
    return true;
  } else if (c == ' ') {
    gs->simulation_active = !gs->simulation_active;
  } else if (c == 'n') {
    bool* b = bool_matrix_at(gs->m_cur, gs->cursor_pos.x, gs->cursor_pos.y);
    *b = !*b;
  } else if (c == 'h' || c == 'a') {
    if (gs->cursor_pos.x > 0) {
      gs->cursor_pos.x--;
    }
  } else if (c == 'l' || c == 'd') {
    if (gs->cursor_pos.x < gs->field_size.x - 1) {
      gs->cursor_pos.x++;
    }
  } else if (c == 'k' || c == 'w') {
    if (gs->cursor_pos.y > 0) {
      gs->cursor_pos.y--;
    }
  } else if (c == 'j' || c == 's') {
    if (gs->cursor_pos.y < gs->field_size.y - 1) {
      gs->cursor_pos.y++;
    }
  } else if (c == '+') {
    if (gs->speed > 1) {
      gs->speed /= 2;
    }
  } else if (c == '-') {
    if (gs->speed < 64) {
      gs->speed *= 2;
    }
  } else if (c == 'c') {
    for (size_t x = 0; x < bool_matrix_width(gs->m_cur); ++x) {
      for (size_t y = 0; y < bool_matrix_height(gs->m_cur); ++y) {
        *bool_matrix_at(gs->m_cur, x, y) = false;
      }
    }
  } else if (c == 'v') {
    gs->selection_active = !gs->selection_active;
    if (gs->selection_active) {
      gs->selection_start = gs->cursor_pos;
    }
  } else if (c == 'u') {
    save_selection(gs, "save_u.gol");
  } else if (c == 'U') {
    load_selection(gs, "save_u.gol");
  } else if (c == 'i') {
    save_selection(gs, "save_i.gol");
  } else if (c == 'I') {
    load_selection(gs, "save_i.gol");
  } else if (c == 'o') {
    save_selection(gs, "save_o.gol");
  } else if (c == 'O') {
    load_selection(gs, "save_o.gol");
  }
  return false;
}

void update(GameState* gs) {
  if (gs->simulation_active && gs->time_step % gs->speed == 0) {
    simulate(gs->m_cur, gs->m_next);
    swap(&gs->m_cur, &gs->m_next);
  }
}

void draw(GameState* gs) {
  // Draw Game Matrix
  for (size_t x = 0; x < gs->field_size.x; ++x) {
    for (size_t y = 0; y < gs->field_size.y; ++y) {
      if (*bool_matrix_at(gs->m_cur, x, y)) {
        Cell* cell = tui_cell_at(x, y);
        cell->content = '#';
      }
    }
  }

  if (!gs->selection_active) {
    // Draw Cursor
    Cell* cell = tui_cell_at(gs->cursor_pos.x, gs->cursor_pos.y);
    cell->text_color = FG_BLACK;
    cell->background_color = BG_WHITE;
  } else {
    // Draw Selection
    Size2 min = min2(gs->cursor_pos, gs->selection_start);
    Size2 max = max2(gs->cursor_pos, gs->selection_start);
    for (size_t x = min.x; x <= max.x; ++x) {
      for (size_t y = min.y; y <= max.y; ++y) {
        Cell* cell = tui_cell_at(x, y);
        cell->text_color = FG_BLACK;
        cell->background_color = BG_WHITE;
      }
    }
  }
}
